function TalentSwitch.rev_custom(active, talent, body)
	-- need recursion protection

	body=body:gsub("#TSW#","")

	local specGroup = GetActiveSpecGroup(); 
	if active == talent then
		return body
	end

	-- reverse old mappings
	if custom[specGroup][talent] ~= nil then

		local sub_list = {}
		for i = 1, #custom[specGroup][talent] do
			local a = TalentSwitch.esc(custom[specGroup][talent][i][1])
			local b = TalentSwitch.esc(custom[specGroup][talent][i][2])

			sub_list[b] = a -- ensures uniqueness of substitution
		end

		for b,a in pairs(sub_list) do

			-- this is to stop recursion when the destination string contains the starting string
				local rec = a:find(b)
				if rec ~= nil then
					body=body:gsub(a,"#TSW#") -- remove already subbed in strings
				end

			body=body:gsub(b,a)

				if rec ~= nil then
					body=body:gsub("#TSW#",a) -- put them back now that the sub has occured
				end
		end

	end

	return body
end

function TalentSwitch.do_custom(active, body)
	-- do new mappings
	local specGroup = GetActiveSpecGroup(); 

	if custom[specGroup][active] ~= nil then

		local sub_list = {}
		for i = 1, #custom[specGroup][active] do
			local a = TalentSwitch.esc(custom[specGroup][active][i][1])
			local b = TalentSwitch.esc(custom[specGroup][active][i][2])
	
			sub_list[a] = b
		end

		for a,b in pairs(sub_list) do
			-- this is to stop recursion when the destination string contains the starting string
				local rec = b:find(a)
				if rec ~= nil then
					body=body:gsub(b,"#TSW#") -- remove already subbed in strings
				end

			body=body:gsub(a,b)

				if rec ~= nil then
					body=body:gsub("#TSW#",b) -- put them back now that the sub has occured
				end
		end
	end

	return body
end

-- custom dub slash commands

function SlashCmdList.TSWITCHCUSTOM(msg, editbox)
if msg ~= nil and msg ~= "" then

		local specGroup = GetActiveSpecGroup();
		if editbox == "TSwitchInternal1" then
			specGroup = 3-specGroup
		end

		if editbox ~= "TSwitchInternal0" and editbox ~= "TSwitchInternal1" then
			msg = msg:gsub("\\n", "\n")
		end

		local arglist = {TalentSwitch.split(msg, "&")}
		if arglist[1] ~= nil and arglist[1] ~= "" and arglist[2] ~= nil and arglist[2] ~= "" and arglist[3] ~= nil and arglist[3] ~= "" then
			if custom[specGroup][arglist[1]] == nil then
				custom[specGroup][arglist[1]] = {}
			end

			-- sanity check
			local left = arglist[2]
			local right = arglist[3]
			if left:find(right) == nil and right:find(left) == nil then
				table.insert(custom[specGroup][arglist[1]], {arglist[2], arglist[3]})
				if editbox ~= "TSwitchInternal0" and editbox ~= "TSwitchInternal1" then
					print("TalentSwitch: Added custom substitution "..arglist[2].." -> "..arglist[3].." for "..arglist[1])
				end
			else
				if editbox ~= "TSwitchInternal0" and editbox ~= "TSwitchInternal1" then
					print("TalentSwitch: Invalid Rule! Consider using a hashtag in the\nshorter macro so that the longer macro does not contain the shorter.")
				else
					TSwitchUIframe.guideText.text:SetText("Invalid Rule! Consider using a hashtag in the\nshorter macro so that the longer macro does not\ncontain the shorter.")
				end
			end

		end
	end
end

function SlashCmdList.TSWITCHLIST(msg,frame)
	local specGroup = GetActiveSpecGroup(); 
	local arglist = {}

	print("TalentSwitch Custom Subsitution List:")

	if msg ~= nil and msg ~= "" then
		arglist = {TalentSwitch.split(msg, "&")}
	else
		local n = 0
		for k,v in pairs(custom[specGroup]) do
			n=n+1
			arglist[n]=k
		end
	end

	for i = 1,#arglist do
		print("Custom substitutions for \'"..arglist[i].."\':")

		if custom[specGroup][arglist[i]] ~= nil and custom[specGroup][arglist[i]][1] ~= nil then
			for j = 1,#custom[specGroup][arglist[i]] do
				print("      ["..j.."] "..custom[specGroup][arglist[i]][j][1].." -> "..custom[specGroup][arglist[i]][j][2])
			end
		else
			print("--None--")
		end
	end

end

function SlashCmdList.TSWITCHDEL(msg,frame)
	local specGroup = GetActiveSpecGroup(); 
	local arglist = {}

	if frame == "TSwitchInternal1" then
		specGroup = 3-specGroup
	end


	if msg~= nil and msg ~= "" then
		arglist = {TalentSwitch.split(msg, "&")}

		if custom[specGroup][arglist[1]] ~= nil and custom[specGroup][arglist[1]] ~= "" then
			table.remove(custom[specGroup][arglist[1]], arglist[2])

			if frame ~= "TSwitchInternal0" and frame ~= "TSwitchInternal1" then
				print("TalentSwitch: Removed "..arglist[2].." from custom substitutions for "..arglist[1])
			end
		else
			if frame ~= "TSwitchInternal0" and frame ~= "TSwitchInternal1" then
				print("TalentSwitch: Nothing to remove")
			end
		end

		if #custom[specGroup][arglist[1]] == 0 then
			custom[specGroup][arglist[1]] = nil
			if frame == "TSwitchInternal0" or frame == "TSwitchInternal1" then
				TalentSwitch.talentSelect(arglist[1], 3)
			end
			return
		end
	end

	if frame == "TSwitchInternal0" or frame == "TSwitchInternal1" then
		if #custom[specGroup][arglist[1]] == 1 then
			TalentSwitch.talentSelect(arglist[1], 0)
		else
			TalentSwitch.talentSelect(arglist[1], 2)
		end
	end
end

-- Utility functions

function TalentSwitch.esc(x)
  return (x:gsub('%%', '%%%%')
           :gsub('%^', '%%%^')
           :gsub('%$', '%%%$')
           :gsub('%(', '%%%(')
           :gsub('%)', '%%%)')
           :gsub('%.', '%%%.')
           :gsub('%[', '%%%[')
           :gsub('%]', '%%%]')
           :gsub('%*', '%%%*')
           :gsub('%+', '%%%+')
           :gsub('%-', '%%%-')
           :gsub('%?', '%%%?'))
end

function TalentSwitch.split(str, c)
	local a = string.find(str, c)
	local aCount = 1
	local start = 1
	local array = {}
	
	while a do
		array[aCount] = string.sub(str, start, a - 1)
		array[aCount] = array[aCount]:gsub("^%s*(.-)%s*$", "%1")
		start = a

		str = string.gsub(str, c, "", 1)
		a = string.find(str, c)
		
		aCount = aCount + 1
	end
	array[aCount] = string.sub(str, start)
	array[aCount] = array[aCount]:gsub("^%s*(.-)%s*$", "%1")

	return unpack(array)
end
